require 'open-uri'
require 'nokogiri'
require 'net-ldap'

module ZmLdifExport

  class << self
    @options = nil
    @file_path = nil

    def ldif_file
      begin
        open @file_path
      rescue Errno::ENOENT => _
        puts "File #{@file_path} not found"
      end
    end

    # Load the LDIF in Memory
    def ldif_data
      ldif_data = Net::LDAP::Dataset.read_ldif ldif_file
      puts "File #{@file_path} not a valid LDIF file" if ldif_data.empty?
      ldif_data
    end

    def parse_ldif
      ldif_data.to_entries.each do |entry|
        Domain.add Domain.new(entry) if entry.objectclass.include?(OCLASS_DOMAIN)
        DistributionList.add DistributionList.new(entry) if entry.objectclass.include?(OCLASS_DL)
        Account.add Account.new(entry) if valid_account_entry?(entry)
      end
    end

    def valid_account_entry?(entry)
      entry.objectclass.include?(OCLASS_ACCOUNT) && entry.dn =~ /(#{domain_to_ldif_dc}$|uid=admin|uid=spam|uid=virus|uid=nospam)/
    end

    # Entry point do all
    def run(options)
     @options = options
     @file_path = @options[:ldif_file]
     @domain = @options[:domain]
     @command = @options[:command]

     parse_ldif
     self.send(@command)
    end

    def exportAccounts
      puts "export LC_ALL=en_US.UTF-8"
      puts "export LANG=en_US.UTF-8"
      Account.all.each do |a|
        next if a.name.first =~ /(admin|spam|ham|galsync|virus|wiki)/
        puts "# zmprov ca #{a.name(:zmprov, @domain)}"
        puts a.zmprov_ca(domain: @domain)
        puts ""
      end
    end

    def exportAliases
      Account.all.each do |a|
        next unless a.has_alias?
        puts "# zmprov aaa #{a.name(:zmprov, @domain)}"
        puts a.zmprov_aaa(domain: @domain)
        puts ""
      end
    end

    def exportDls
      puts "export LC_ALL=en_US.UTF-8"
      puts "export LANG=en_US.UTF-8"
      DistributionList.all.each do |dl|
        next if dl.name(:zmprov) =~ /archive/i
        puts "# Create DistributionList #{dl.name(:zmprov, @domain)}"
        puts dl.zmprov_cdl domain: "@#{@domain}"
        puts ""
        puts "# Add members to DistributionList #{dl.name(:zmprov, @domain)}"
        puts dl.zmprov_adlm domain: "@#{@domain}"
        puts ""
        puts "# Add Alias to DistributionList #{dl.name(:zmprov, @domain)}"
        puts dl.zmprov_adla domain: "@#{@domain}"
        puts ""
      end
    end

    def domain_to_ldif_dc
      "dc=#{@domain.split('.').join(',dc=')}"
    end

    # Build the resulting command
    # @param zm_cmd [String] `zmprov` or `zmmailbox`
    # @param command [String] command to excute, like `ca` or `createDomain`
    # @param attributes [Array] Array of attributes to pass to the command
    # @param full [Boolean] If print `zm_cmd` or just the `command`
    # @returns [String]
    def mk_command(zm_cmd: 'zmprov', command:, resource:, attributes: {}, full: true)
      comment = "# #{zm_cmd} #{command} #{resource}"
      command = full ? "#{zm_cmd} #{command} #{resource}" : "#{command} #{resource}"
      command = command + " \\" if attributes.keys.any?
      attributes = attrs_to_cmd(attributes)
      
      [comment, command, attributes].compact.join("\n")
    end

    # Return an array of attributes
    def attrs_to_cmd(attributes = {})
      lines = []
      attributes.each do |k,v|
        values = [v].flatten.compact
        values.each do |value|
          lines.push "#{k} #{value} \\"
        end
      end
      return nil if lines.empty?
      lines.last.gsub!(' \\', '')
      return lines
    end

  end

  # CONSTANTS
  ZIMBRA_ATTRS_FILE = './data/zimbra-attrs.xml'.freeze
  LDIF_FILE = ARGV[0]
  OCLASS_DOMAIN = 'zimbraDomain'
  OCLASS_ACCOUNT = 'zimbraAccount'
  OCLASS_DL = 'zimbraDistributionList'

#  LDIF_DATA = Net::LDAP::Dataset.read_ldif(open(LDIF_FILE)).to_entries
  @domains = nil
  @accounts = nil
  @aliases = nil

  class Attribute
    attr_reader :name, :type

    WHITE_LIST = ['zimbraId', 'zimbraMailAlias'].freeze

    def initialize(xml_attr)
      @name = xml_attr.attributes['name'].value
      @singular = attr_is_singular?(xml_attr)
      @immutable = attr_is_immutable?(xml_attr)
      @type = xml_attr.attributes['type'].value
    end

    def singular?
      @singular
    end

    def immutable?
      @immutable
    end

    def multiline?
      name == 'zimbraPrefMailSignatureHTML' || name == 'zimbraMailSieveScript' || name == 'zimbraPrefOutOfOfficeReply' || name == 'zimbraPrefOutOfOfficeExternalReply'
    end

    def text?
      type == 'string' || type == 'duration' || name == 'zimbraACE' || type == 'regex'
    end

    def number?
      type == 'integer' || type == 'long'
    end

    def enum?
      type == 'enum'
    end

    def boolean?
      type == "boolean"
    end

    def to_zmprov(value, plus = false)
      if multiline?
        command_multiline attr_name: name, value: value
      elsif singular?
        return " \\\n#{name} '#{value.join('').to_s}'" if text?
        return " \\\n#{name} #{value.join('').to_i}" if number?
        return " \\\n#{name} #{value.join('').to_s}"
      else
        line = " \\\n#{name} "
        value = text? ? value.map {|v| "'#{v}'"} : value
        if plus
          line << value.join(" +#{name} ")
        else
          line << value.join(" \\\n#{name} ")
        end
          
        line
      end
    end

    def command_multiline(attr_name:,value:)
      final_wrapper = attr_name == 'zimbraMailSieveScript' ? "'" : '"'
      original_wrapper = attr_name == 'zimbraMailSieveScript' ? '"' : "'"
      line = "#{attr_name} #{final_wrapper}"
      line << value.join('').to_s.gsub(final_wrapper, original_wrapper).split(/\n/).join("\n")
      line << "#{final_wrapper} \\\n"
      line
    end

    private

    def attr_is_singular?(attr)
      return false if attr.attributes['cardinality'].nil?
      return true if attr.attributes['type'] == 'astring'
      attr.attributes['cardinality'].value == 'single'
    end

    def attr_is_immutable?(attr)
      return false if WHITE_LIST.include?(name)
      return false if attr.attributes['immutable'].nil? 
      attr.attributes['immutable'].value == '1'
    end
  end

  class ZimbraResource
    
    BLACK_LIST_ATTRS = [
      "dn", "objectclass", "o", "dc", "structuralobjectclass", "entryuuid", "creatorsname",
      "createtimestamp", "entrycsn", "modifiersname", "modifytimestamp", "uid",
      "zimbraaggregatequotalastusage", "zimbradomainstatus", "zimbramailstatus", "zimbraauthtokens",
      "amavisarchivequarantineto", 'zimbramailhost', 'zimbramailtransport', 'zimbramailquota',
      'zimbraaccountstatus', 'zimbraarchiveaccount', 'zimbraarchiveenabled', 'zimbraauthtokenvalidityvalue',
      'zimbraPasswordModifiedTime', 'zimbrapreftimezoneid', 'zimbraquotalastwarntime', 'zimbracsrftokendata',
      'zimbraprefpop3downloadsince', 'zimbrapasswordmodifiedtime', 'zimbracosid',
      'zimbrapreflistviewcolumns', 'zimbraisadminaccount', 'zimbraisadminaccount', 'zimbraAdminSavedSearches',
      'zimbraisadmingroup', 'zimbraadminconsoleuicomponents'
    ]

    attr_reader :attributes, :name, :real_attrs

    @immutable_attrs = nil
    @attrs_xml = nil

    def self.attrs_xml
      return @attrs_xml if @attrs_xml
      xml_doc = Nokogiri::XML(open(ZIMBRA_ATTRS_FILE))
      @attrs_xml = xml_doc.xpath('//attr')
      return @attrs_xml
    end

    def zimbra_attrs
      return @zimbra_attrs if @zimbra_attrs
      @zimbra_attrs = {}
      ZimbraResource.attrs_xml.each do |attr|
        zm_attr = Attribute.new(attr)
        @zimbra_attrs[zm_attr.name.downcase] = zm_attr
      end
      return @zimbra_attrs
    end

    def name(mode = nil, domain = '')
      return @name if mode.nil?
      return "#{@name.join('').to_s}@#{domain}" if mode == :zmprov
    end

    def id
      attributes['zimbraid'].first.to_s
    end

    def parse_attributes(ldif:, exclude: [])
      attrs = {}
      ldif.each_attribute do |a|
        value = ldif[a]
        next if invalid_attribute?(a.to_s) || exclude.include?(a) || invalid_value?(value)
        attrs[a.to_s] = value
      end
      attrs
    end

    def invalid_value?(value)
      return true if value.join('').to_s =~ /null/
    end

    def invalid_attribute?(attr_name)
      return true if BLACK_LIST_ATTRS.include?(attr_name)
      zm_attr = zimbra_attrs[attr_name]
      zm_attr.nil? ? false : zm_attr.immutable?
    end

    def to_zmprov(command: 'ca')
      zmprov_ca if command == 'ca'
      zmprov_aaa if command == 'aaa'
      zmprov_cdl if command == 'cdl'
      zmprov_adlm if command == 'adlm'
      zmprov_adla if command == 'adla'
    end

    def attr_to_zmprov(attr_name:, value:, plus: false)
      zm_attr = zimbra_attrs[attr_name]
      zm_attr.to_zmprov(value, plus)
    end
  end

  class Domain < ZimbraResource
    @resources = []

    def self.all
      return @resources
    end

    def self.add(resource)
      @resources.push resource
    end

    def initialize(ldif)
      @name = ldif[:zimbradomainname]
      @attributes = parse_attributes(ldif: ldif, exclude: [:zimbradomainname])
    end

    def zmprov_cd
      line = ['cd', name]
      attributes.each do |name, value|
        line << attr_to_zmprov(attr_name: name, value: value)
      end
      line.join(' ')
    end

  end

  class Account < ZimbraResource
    @resources = []
    attr_accessor :ldif

    def self.all
      return @resources
    end

    def self.add(resource)
      @resources.push resource
    end

    def has_alias?
      attributes['zimbramailalias'] ? attributes['zimbramailalias'].any? : false
    end

    def initialize(ldif)
      @name = ldif[:uid]
      @ldif = ldif
      @attributes = parse_attributes(ldif: ldif, exclude: [:uid])
    end


    def zmprov_ca(domain: '', mail_host: nil, skip: [], print: false)
      line = [ 
        '/opt/zimbra/bin/zmprov ca',
        name(:zmprov, domain),
        'P4ssW04r..,. '
      ]
      line.push "zimbraMailHost #{mail_host} " unless mail_host.nil?
      attributes.each do |name, value|
        next if name =~ /(zimbramailalias|admin|zimbracosid|zimbraace|zimbrasharedItem|signature|zimbramailsievescript)/i
        next if skip.include?(name)
        line << attr_to_zmprov(attr_name: name, value: value)
      end
      cmd = line.join(' ')
      puts cmd if print
      return cmd
    end

    def zmprov_aaa(print: false, domain:)
      aliases = attributes['zimbramailalias']
      return if aliases.nil? || aliases.empty?
      lines = []
      aliases.each do |mail|
        lines.push("aaa #{name(:zmprov, domain)} #{mail}")
      end
      
      cmd = lines.join("\n")
      puts cmd if print
      return cmd
    end

    def ace_to_json(zm_attr = '')
      ace = {}
      keys_and_values = zm_attr.split(';')
      keys_and_values.each do |kv|
        key, value = kv.split(':')
        ace[key] = kv[value]
      end
      return ace
    end

    def aces(zma)
      aces = []
      zma.each do |zm|
        aces.push(ace_to_json(zm))
      end

      return aces
    end

    def zmprov_sharing(print: false, domain:)
      return nil if attributes['zimbrashareditem'].nil?
      line = []
      zimbra_shares = aces(attributes['zimbrashareditem'])
      zimbra_shares.each do |ace|
        next if ace['granteeName'] == 'null'
        line.push("zmmailbox -z -m #{name(:zmprov, domain)} mfg '#{ace['folderPath']}' account #{ace['granteeName']} #{ace['rights']}")
      end

      cmd = line.size > 0 ? line.join("\n") + "\n" : nil
      puts cmd if print
      return cmd
    end

    def zmprov_signatures(print: false, domain:)
      return nil if attributes['zimbraprefmailsignaturehtml'].nil?
      line = []
      zimbra_shares = attributes['zimbraprefmailsignaturehtml']
      zimbra_shares.each do |ace|
        ace = ace.gsub(/'/, '"')
        line.push("ma #{name(:zmprov, domain)} zimbraPrefMailSignatureHtml '#{ace}'")
      end

      cmd = line.size > 0 ? line.join("\n") + "\n" : nil
      puts cmd if print
      return cmd
    end
    
  end

  class DistributionList < ZimbraResource
    @resources = []

    def self.all
      return @resources
    end

    def self.add(resource)
      @resources.push resource
    end

    def initialize(ldif)
      @name = ldif[:uid]
      @attributes = parse_attributes(ldif: ldif, exclude: [:uid])
    end

    def zmprov_cdl(domain: '', print: false)
      line = ['zmprov cdl', name(:zmprov, domain)]
      attributes.each do |name, value|
        next if name =~ /(zimbraMailForwardingAddress)|(zimbraMailAlias)/i
        line << attr_to_zmprov(attr_name: name, value: value)
      end
      cmd = line.join(' ')
      puts cmd if print
      return cmd
    end

    def zmprov_adlm(domain:, print: false)
      line = ['adlm', name(:zmprov, domain)]
      attributes.each do |name, value|
        next unless name =~ /zimbraMailForwardingAddress/i
        line << value.join(' ')
      end
      cmd = line.join(' ')
      puts cmd if print
      return cmd
    end

    def zmprov_adla(domain:, print: false)
      lines = []
      attributes.each do |name, value|
        next unless name =~ /zimbraMailAlias/i
        value.each do |v|
          lines << "adla #{name(:zmprov, domain)} #{v}" 
        end
      end
      cmd = lines.join(' ')
      puts cmd if print
      return cmd
    end
  end




  # Domain.all.each do |domain|
  #   next if domain.name(:zmprov) =~ /archive/i
  #   puts domain.zmprov_cd
  #   puts "\n"
  # end

  
end

